from iceberg_tools.schema.simplified_validator import ensure_schema


def test_cytoscape(distribution_schema_path):
    """Ensure specimen and document_reference have _identifier denormalized fields."""
    schema = ensure_schema(schema_path=distribution_schema_path)
    assert schema, f"should have loaded {distribution_schema_path}"
    ignore = ['root', 'data_release', 'core_metadata_collection']
    sources = [_ for _ in schema.keys() if _ not in ignore]

    edge_table_path = "/tmp/simplified_network_table.tsv"
    with open(edge_table_path, "w") as fp:
        print("source\ttarget\tname", file=fp)
        for source in sources:
            for _ in schema[source]['links']:
                assert 'target_type' in _, f"should have target in {source} {_}"
                assert 'name' in _, f"should have name in {source} {_}"

            targets = [(_['target_type'], _['name']) for _ in schema[source]['links']]
            for target, name in targets:
                print(f"{source}\t{target}\t{name}", file=fp)





def test_schema_load(fhir_schema, fhir_validator, vertex_schemas, edge_schemas):
    """Ensure whole schema can load"""
    assert 'DocumentReference' in fhir_schema['$defs']
    document_reference = fhir_schema['$defs']['DocumentReference']
    assert document_reference
    link_hrefs = set([_['href'].split('/')[0] for _ in document_reference['links']])
    out_of_scope = set([
        'RelatedPerson',
        'Practitioner',
        'PractitionerRole',
        'Organization'])
    assert (out_of_scope - link_hrefs) == out_of_scope, f"Did not expect {out_of_scope - link_hrefs}"



import subprocess


def test_lathe_lint(bmeg_dir, lath_path):
    """Run lathe's schema-lint command."""
    cmd = f"{lath_path} schema-lint .".split()
    results = subprocess.run(cmd, capture_output=True, text=True)
    assert results.returncode == 0, (results.stdout, results.stderr)
    assert 'ERROR' not in results.stdout, results.stdout
    lines = results.stdout.split('\n')
    for _ in lines:
        if len(_) > 0:
            assert 'OK' in _, _



import importlib
import json
from string import Formatter
from typing import List, Iterator, Callable
from glom import glom, PathAccessError, flatten

import fastjsonschema
import jsonschema
import requests
import re
from jsonpointer import resolve_pointer

from iceberg_tools.schema import extract_schemas, BASE_URI

NESTED_OBJECTS_IGNORE = ['Identifier', 'Extension']


class RefFinder:
    """
    For returning a list of paths to schema refs

    Parameters:
        schema: the json schema object to be traversed
        refs: a flag that decides wether Reference.yaml refs will be collected
                or if every other ref will be collected
    Returns:
        list: A list of glom paths to schema refs
    """
    def __init__(self, schema, refs, nested_objects=None):
        self.schema = schema
        self.refs = refs
        self.nested_objects = nested_objects or []

    def getpath(self, data, path):
        """Reassign schema data dict to the new index path of every key in new_path"""
        for key in path:
            data = data[key]
        return data

    def traverse(self, path, data):
        """Traverse through a schema collecting a list of glom $ref paths"""
        if isinstance(data, dict):
            for key, value in data.items():
                new_path = path + [key]
                if key == "$ref":
                    ref_value = self.getpath(self.schema, new_path[:-1])[key]
                    if (ref_value != "Reference.yaml" and not self.refs) or (ref_value == "Reference.yaml" and self.refs):
                        self.nested_objects.append(".".join(map(str, new_path)))
                self.traverse(new_path, value)
        elif isinstance(data, list):
            for index, value in enumerate(data):
                new_path = path + [index]
                self.traverse(new_path, value)

    def find_refs(self):
        """Calling function kickcs off schema traversal"""
        self.traverse([], self.schema)
        return self.nested_objects


def _extract_target_hints(schema_link):
    """Retrieve attributes from targetHints, using defaults."""
    target_hints = schema_link.get('targetHints', {})
    multiplicity = next(iter(target_hints.get('multiplicity', [])), 'has_many')
    directionality = next(iter(target_hints.get('directionality', [])), 'one')
    association = next(iter(target_hints.get('association', [])), False)
    return directionality, multiplicity, association


def _generate_links_from_fhir_references(schema, classes) -> List[dict]:
    """Generate links for a schema.

    Parameters
        schema: a schema dict
        classes: a list of types
    Returns:
        tuple: (links, nested_links)
    """

    # Direct links from {schema['title']}"
    links = []
    links.extend(_extract_links(schema, classes))

    # Nested links
    nested_links = []
    for nested_schema, path in _extract_nested_schemas(schema):
        if nested_schema['title'] in NESTED_OBJECTS_IGNORE:
            continue
        extracted_links = _extract_links(nested_schema, classes)
        if len(extracted_links) == 0:
            continue

        path_parts = path.split('.')
        parent_path = '.'.join(path_parts[1:-1]).replace('.items', '.-')

        for _ in extracted_links:
            _['$comment'] = f"From {nested_schema['title']}/{parent_path.replace('.-', '')}"
            _['templatePointers'] = {"id": f"/{parent_path.replace('.', '/')}{_['templatePointers'][tp]}" for tp in _['templatePointers']}
            _['rel'] = f"{parent_path.replace('.-', '')}_{_['rel']}"
        nested_links.extend(extracted_links)

    return links, nested_links


def _extract_nested_schemas(schema) -> Iterator[tuple[dict, str]]:
    """Extract $ref from elements in schema that are not `References`.

    Returns: (sub_schema, path) a tuple of the sub schema and the path to it in the passed schema
    """
    refs_finder = RefFinder(schema, refs=False)
    matches = sorted(refs_finder.find_refs())

    mod = importlib.import_module('fhir.resources')
    for match in matches:
        if match.startswith('properties.links') or match.startswith('links'):
            continue
        property_name = match.split('.')[1]
        if property_name.startswith('_'):
            continue

        property_ = schema['properties'][property_name]

        multiplicity = 'has_one'
        if 'items' in property_:
            multiplicity = 'has_many'

        if multiplicity == 'has_many':
            target = property_['items']['$ref'].split('.')[0]
        else:
            assert '$ref' in property_, ("TODO no $ref?", property_name, property_, match)
            target = property_['$ref'].split('.')[0]

        target_class = mod.get_fhir_model_class(target)
        sub_schemas = extract_schemas([target_class], BASE_URI)
        sub_schema = sub_schemas[target]

        yield sub_schema, match


def _extract_links(schema: dict, classes) -> List[dict]:
    """Extract Link Description Object (LDO) from a schema.

    see https://json-schema.org/draft/2019-09/json-schema-hypermedia.html#rfc.section.6
    """
    refs_finder = RefFinder(schema, refs=True)
    matches = sorted(refs_finder.find_refs())

    # filter out only the matches that link property references
    matches = [match for match in matches if match.startswith("properties.")]
    links = []
    for match in matches:
        direction = 'outbound'
        multiplicity = 'has_one'
        if 'items' in match:
            multiplicity = 'has_many'
        property_name = match.split('.')[1]
        property_ = schema['properties'][property_name]
        if 'enum_reference_types' not in property_:
            property_['enum_reference_types'] = ['__ANY__']
        append_postscript = len(property_['enum_reference_types']) > 1
        _path = '.'.join(match.split('.')[1:-1])
        _path = _path + '.reference'
        _path = _path.replace('.items', '.-')

        _class_names = [_.__name__ for _ in classes]

        for enum_reference_type in property_['enum_reference_types']:
            if enum_reference_type not in _class_names:
                print(f"DEBUG: {enum_reference_type} not in scope, skipping link {schema['$id']}->{enum_reference_type}")
                continue
            rel = f"{property_name}"
            if append_postscript:
                rel += f"_{enum_reference_type}"
            links.append(
                {
                    "rel": rel,
                    "href": f"{enum_reference_type}/{{id}}",
                    "templateRequired": ["id"],
                    "targetHints": {
                        "multiplicity": [multiplicity],
                        "direction": [direction],
                        'backref': [property_['backref']],
                    },
                    "templatePointers": {
                        'id': "/" + _path.replace('.', '/')
                    },
                    'targetSchema': {'$ref': enum_reference_type + '.yaml'},
                }
            )
    return links


def _load_schema(schema) -> dict:
    """Load JSON schema from dict or string."""
    if isinstance(schema, str):
        schema_ = requests.get(schema).json()
    elif isinstance(schema, dict):
        schema_ = schema
    else:
        raise ValueError("Schema must be a URL or a dict")
    return schema_


def fastjsonschema_compile(schema: dict, formats: dict = {}) -> Callable:
    """Compile a JSON schema, explicitly provide handlers."""

    def _handler(uri):
        """Resolve $ref to $id"""
        for _def in schema['$defs'].values():
            if _def['$id'] == uri:
                print(f"Found {uri}")
                return _def
        print(f"Could not find {uri}")

    def _https_handler(uri):
        # ‘https://json-schema.org/draft/2020-12/links’  hosted by cloudflare
        # see  https://github.com/IATI/IATI-Standard-Website/issues/230
        import urllib.request
        req = urllib.request.Request(
            uri,
            data=None,
            headers={
                'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.1916.47 Safari/537.36'
            }
        )
        f = urllib.request.urlopen(req)
        _ = f.read().decode('utf-8')
        return json.loads(_)

    compiled_schema = fastjsonschema.compile(
        schema,
        formats=formats,
        handlers={'http': _handler, 'https': _https_handler},
    )
    return compiled_schema


class AssociationSchema:
    """A JSON schema for an association."""

    def __init__(self, schema):
        """Load and compile a JSON schema."""
        if isinstance(schema, str):
            self.schema = requests.get(schema).json()
        elif isinstance(schema, dict):
            self.schema = schema
        else:
            raise ValueError("Schema must be a URL or a dict")
        self.compiled_schema = fastjsonschema_compile(self.schema)

    def validate(self, instance: dict) -> dict:
        """Validate data against schema.

        Args:
            instance: instance of data to validate
        Returns:
            dict: potentially modified instance
        """
        return self.compiled_schema(instance)

    @classmethod
    def is_association(cls, schema: dict) -> bool:
        """Is this schema an association?"""
        if 'links' not in schema:
            return False
        for _ in schema['links']:
            _, _, association = _extract_target_hints(_)
            if association:
                return True
        return False

    @classmethod
    def validate_schema_conventions(cls, schema: dict) -> bool:
        """Validate schema."""
        assert 'links' in schema, 'Schema should have links'
        assert len(schema['links']) >= 2, "Association schema should have at least two links"
        assert cls.is_association(schema), "Schema links should be an association"
        assert 'title' in schema, "Schema should have a title"
        return True

    def validate_links(self, instance: dict):
        """Validate instance.links against schema.links"""

        # verify that all links in instance are in schema
        schema_relationships = [_['rel'] for _ in self.schema['links']]
        for link in instance['links']:
            assert link['rel'] in schema_relationships, f"Instance of this association should have links to {schema_relationships}"

        return instance


class AssociationInstance:

    def __init__(self, association_schema: AssociationSchema, instance: dict = None, vertex_a: dict = None, vertex_b: dict = None):
        """Operate on json document in the context of an association schema.

        Parameters:
            association_schema: an AssociationSchema object
            instance: an instance of a JSON document that should conform to the association schema
            vertex_a: a vertex that should be associated
            vertex_b: a vertex that should be associated
        """
        self.instance = instance
        self.association_schema = association_schema
        self.regexp_cache = {}
        self.href_keys_cache = {}
        if instance is None:
            assert vertex_a is not None and vertex_b is not None, "Must provide instance or vertex_a and vertex_b"
            self.instance = self._create_instance(vertex_a, vertex_b)

    def __repr__(self):
        """Return a string representation of the instance links of the form V(id).backref<--Title-->V(id).backref"""
        parts = self.edge_parts()
        label = parts['label']

        # formulate a representation of the edge's vertices
        rels = [_ for _ in parts if _ != 'label']
        msgs = []
        for rel in rels:
            _ = parts[rel]
            msgs.append(f"{_['targetSchema']}({_[_['id_name']]}).{rel}")

        return f"{msgs[0]}<-{label}->{msgs[1]}"

    def _create_instance(self, vertex_a, vertex_b):
        """Create an instance of an association from two vertices."""
        links = []
        for schema_link, vertex in list(zip(self.association_schema.schema['links'], [vertex_a, vertex_b])):
            keys = self._extract_href_keys(schema_link['href'])
            values = [resolve_pointer(vertex, _) for _ in schema_link['templatePointers'].values()]
            links.append(
                {
                    'rel': schema_link['rel'],
                    'href': schema_link['href'].format(**dict(zip(keys, values))),
                }
            )
        return {
            'links': links
        }

    def _extract_href_keys(self, href):
        """Extract the keys from the href template, cache the result."""
        keys = self.href_keys_cache.get(href, None)
        if not keys:
            keys = [_[1] for _ in Formatter().parse(href)]
            self.href_keys_cache[href] = keys
        return keys

    def _extract_id(self, schema_href, instance_href) -> str:
        """Read the id from the href template."""
        rexp = self.regexp_cache.get(schema_href, None)
        if rexp is None:
            _ = schema_href
            keys = self._extract_href_keys(schema_href)
            for key in keys:
                _ = _.replace(f"{{{key}}}", '(.*)')
            rexp = re.compile(_)
            self.regexp_cache[schema_href] = rexp
        m = rexp.match(instance_href)
        assert m is not None, f"Unable to find id in instance link {instance_href} given schema {schema_href}"
        return m.group(1), self._extract_href_keys(schema_href)[0]

    def edge_parts(self):
        """Decompose the edge into its parts."""

        title = self.association_schema.schema['title']
        parts = {'label': title}

        for link in self.association_schema.schema['links']:
            _ = self._extract_link_parts(link, self.instance)
            parts[_['rel']] = _

        return parts

    def _extract_link_parts(self, schema_link: dict, instance: dict) -> dict:
        """Extract the parts of a link from the schema and instance."""
        instance_link = next(iter([_ for _ in instance['links'] if _['rel'] == schema_link['rel']]), None)

        id_, id_name = self._extract_id(schema_link['href'], instance_link['href'])

        directionality, multiplicity, _ = _extract_target_hints(schema_link)

        return {
            id_name: id_,
            'rel': schema_link['rel'],
            'targetSchema': schema_link['targetSchema']['$ref'],
            'multiplicity': multiplicity,
            'directionality': directionality,
            'id_name': id_name,
        }


class VertexSchemaDecorator:
    """Adds links to vertex schema."""

    def __init__(self, schema: dict, classes: list):
        """Load and compile a JSON schema."""
        self.schema = _load_schema(schema)
        # add links property
        self.schema['properties']['links'] = {
            'links': {
                'type': 'array',
                'items': {
                    '$ref': 'https://json-schema.org/draft/2020-12/links'
                }
            }
        }
        # add links element
        links, nested_links = _generate_links_from_fhir_references(schema, classes)
        self.schema['links'] = links + nested_links
        # check schema
        jsonschema.Draft202012Validator.check_schema(schema)   # Draft202012Validator.check_schema(schema)


def cast_json_pointer_to_glom(_):
    """Convert a JSON pointer to a glom query.
    turn the pointer into a glom expression, note the leading dot and `-` for lists
    refer to https://glom.readthedocs.io/en/latest/api.html#glom.glom
    """
    _ = _.replace('/', '', 1).replace('/', '.').replace('-', '*')
    return _


class VertexLinkWriter:
    """Context manager for inserting links into vertices."""

    def __init__(self, schema: dict):
        """Initialize the context manager.

        Parameters:
            schema: a schema
        """
        self.vertex_schema = schema
        self.href_keys_cache = {}
        self.glom_cache = {}
        self.regexp_cache = {}

    def _extract_href_keys(self, href):
        """Extract the keys from the href template, cache the result."""
        keys = self.href_keys_cache.get(href, None)
        if not keys:
            keys = [_[1] for _ in Formatter().parse(href)]
            self.href_keys_cache[href] = keys
        return keys

    def __enter__(self):
        """Enter the context."""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Exit the context."""
        pass

    def insert_links(self, vertex: dict) -> dict:
        """Insert links into a vertex.

        Parameters:
            vertex: a vertex
        Returns:
            dict: vertex with links inserted

        """
        links = []

        _schema = self.vertex_schema
        if isinstance(self.vertex_schema, VertexSchemaDecorator):
            _schema = self.vertex_schema.schema

        if 'links' not in _schema:
            return vertex

        for schema_link in _schema['links']:

            keys = self._extract_href_keys(schema_link['href'])

            values = self._extract_values(schema_link, vertex)

            if values is None or (isinstance(values, list) and len(values) == 1 and values[0] is None):
                continue

            for _ in values:
                if not isinstance(_, list):
                    _ = [_]
                if 'Resource' in schema_link['href']:  # Any resource
                    links.append(
                        {
                            'rel': schema_link['rel'],
                            'href': _[0],
                        }
                    )
                else:
                    links.append(
                        {
                            'rel': schema_link['rel'],
                            'href': schema_link['href'].format(**dict(zip(keys, _))),
                        }
                    )

        vertex['links'] = links
        return vertex

    def validate(self, vertex: dict):
        """Validate a vertex.

        Parameters:
            vertex: a vertex
        Returns:
            jsonschema.ValidationError if the vertex is invalid, True otherwise

        """
        jsonschema.validate(vertex, self.schema)
        return True

    def _extract_values(self, schema_link, vertex) -> list:
        """Extract values from a vertex given a link description object"""
        values = []
        for k, v in schema_link['templatePointers'].items():
            values_ = self.extract_json_pointer_via_glom(v, vertex)
            if len(values_) == 1:
                values_ = values_[0]

            if values_ is None or (isinstance(values_, list) and None in values_):
                #  Unable to resolve {schema_link['href']} {schema_link['templatePointers']}
                return [None]

            # disambiguate polymorphic references
            # if we have a FHIR polymorphic reference, we need to check that the target schema is in the list of values
            ref = schema_link['targetSchema']['$ref'].split('/')[-1]
            ref = ref.replace('.yaml', '')
            if ref != 'Resource':  # skip `Any` Resource
                if '/' in values_ and ref not in values_:
                    # Polymorphic reference {schema_link['targetSchema']['$ref']} {values} skipping
                    return [None]

            new_val = []
            if isinstance(values_, str):
                values_ = [values_]
            for _ in values_:
                if isinstance(_, list) and len(_) == 1:
                    _ = _[0]

                extracted_value = self._extract_value(schema_link['href'], _)[0]
                new_val.append(extracted_value)
            values.extend(new_val)

        return values

    def extract_json_pointer_via_glom(self, json_pointer, vertex):
        """Convert json pointer to glom query, cache the compiled and extract the value from the vertex. flatten nested lists."""
        if json_pointer not in self.glom_cache:
            glom_instance = cast_json_pointer_to_glom(json_pointer)
            self.glom_cache[json_pointer] = self.glom_cache.get(json_pointer, None) or glom_instance
        try:
            values_ = glom(vertex, self.glom_cache[json_pointer])
        except PathAccessError:
            return [None]
        # flatten nested lists
        if isinstance(values_, list) and len(values_) > 0 and isinstance(values_[0], list):
            values_ = flatten(values_)
        return values_

    def _extract_value(self, schema_href, instance_href) -> (str, str):
        """Read the value from instance href, strips fhir relative references.
        Returns: (instance_value, key)
            """
        # not a fhir relative reference
        if '/' not in instance_href:
            key = self._extract_href_keys(schema_href)[0]
            return instance_href, key

        # cache the regexp
        rexp = self.regexp_cache.get(schema_href, None)
        if rexp is None:
            _ = schema_href
            keys = self._extract_href_keys(schema_href)
            for key in keys:
                _ = _.replace(f"{{{key}}}", '(.*)')
            rexp = re.compile(_)
            self.regexp_cache[schema_href] = rexp

        _ = self._extract_href_keys(schema_href)[0]

        if 'Resource' in schema_href:  # Any Resource
            return instance_href, _

        m = rexp.match(instance_href)
        assert m is not None, f"Unable to find value in instance link {instance_href} given schema {schema_href}"
        return m.group(1), _


class SchemaLinkWriter:
    """Context manager for inserting links into schemas."""

    def __init__(self):
        """Initialize the context manager.
        """
        pass

    def __enter__(self):
        """Enter the context."""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Exit the context."""
        pass

    @staticmethod
    def insert_links(schema, classes) -> dict:
        """Insert links into a schema.

        Parameters:
            schema: a schema dict or url
            classes: a list of types
        Returns:
            dict: schema with links inserted

        """
        schema = _load_schema(schema)
        links, nested_links = _generate_links_from_fhir_references(schema, classes)
        schema['links'] = links + nested_links
        schema['properties']['links'] = {
            'type': 'array',
            'items': {
                '$ref': 'https://json-schema.org/draft/2020-12/links'
            }
        }
        return schema

    @staticmethod
    def validate(schema: dict):
        """Validate a schema.

        Parameters:
            schema: a schema
        Returns:
            jsonschema.ValidationError if the schema is invalid, True otherwise

        """
        jsonschema.Draft202012Validator.check_schema(schema)
        return True



import pathlib

from orjson import orjson

from iceberg_tools.graph import VertexLinkWriter


def test_add_links_to_fhir(fhir_schema):
    """Pluck out the links from the FHIR instance and add them to the graph instance.
    Runtime: 1m """

    # quick lookup for schema
    schema_lookup = {v['title']: v for v in fhir_schema['$defs'].values()}
    emitter_lookup = {v['title']: VertexLinkWriter(v) for v in fhir_schema['$defs'].values()}
    # TODO fhir_schema set CWD to iceberg - fix this
    study_path = '../../../tests/fixtures/simplify/study'
    for path in pathlib.Path(study_path).glob("*.bundle.json"):
        with open(path) as fp:
            # load bundle
            _ = orjson.loads(fp.read())
            for entry in _['entry']:
                # pluck out the resource
                resource = entry['resource']
                resource_type = resource['resourceType']
                # lookup the schema
                schema = schema_lookup.get(resource_type, None)
                assert schema is not None, (resource_type, schema_lookup.keys())
                # lookup the link writer
                link_writer = emitter_lookup.get(resource_type, None)
                assert link_writer is not None, (resource_type, link_writer.keys())
                # add the links
                resource = link_writer.insert_links(resource)
                # validate
                if 'links' in schema:
                    assert 'links' in resource



import json
import logging
import pathlib
from typing import Iterator

import fastjsonschema
import inflection
import orjson

import os
from iceberg_tools.dictionaryutils import DataDictionary
from jsonschema.exceptions import ValidationError

from iceberg_tools.util import _to_file, ParseResult


logger = logging.getLogger(__name__)

LOGGED_ALREADY = []
COMPILED_SCHEMAS = {}


def log_once(msg):
    if msg not in LOGGED_ALREADY:
        logger.info(msg)
        LOGGED_ALREADY.append(msg)


def validate(gen3_resource: dict, schemas: dict):
    """Ensure resource is valid for simplified schema."""

    def _handler(uri):
        """Resolve $ref to $id"""
        for _def in schemas['$defs'].values():
            if _def['$id'] == uri:
                print(f"Found {uri}")
                return _def
        print(f"Could not find {uri}")

    def _https_handler(uri):
        # ‘https://json-schema.org/draft/2020-12/links’  hosted by cloudflare
        # see  https://github.com/IATI/IATI-Standard-Website/issues/230
        import urllib.request
        req = urllib.request.Request(
            uri,
            data=None,
            headers={
                'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.1916.47 Safari/537.36'
            }
        )
        f = urllib.request.urlopen(req)
        _ = f.read().decode('utf-8')
        return json.loads(_)

    try:
        object_ = None
        for expected_property in ['id', 'relations', 'object']:
            assert expected_property in gen3_resource, f"{expected_property} not in {gen3_resource}"

        object_ = gen3_resource['object']
        resource_type = object_['resourceType']
        key = inflection.underscore(resource_type)
        schema = schemas.get(key, schemas.get(f"{key}.yaml", None))
        assert schema, f"Could not find schema for {key}"
        actual_keys = set(object_.keys())
        expected_keys = set(schema['properties'].keys()).union(set([_['name'] for _ in schema['links']]))
        if not actual_keys.issubset(expected_keys):
            if not schema.get('additionalProperties', False):
                assert False, f"Is not a subset {actual_keys - expected_keys} not expected"
            else:
                for _ in actual_keys - expected_keys:
                    msg = f"extra property: {resource_type}.{_} {type(object_[_]).__name__}"
                    log_once(msg)
        if 'required' in schema:
            for _ in schema['required']:
                assert _ in object_, f"{_} missing {object_}"
        compiled_schema = COMPILED_SCHEMAS.get(key, None)
        if not compiled_schema:
            compiled_schema = fastjsonschema.compile(
                schema,
                formats={
                    'time': r'^(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$',
                    'date': r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])?$',
                    'binary': r'^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$',
                    "date-time": r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$',
                    "uri": r'\w+:(\/?\/?)[^\s]+',
                    "uuid": r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-5][0-9a-f]{3}-[089ab][0-9a-f]{3}-[0-9a-f]{12}$"
                },
                handlers={'http': _handler, 'https': _https_handler}
            )
            COMPILED_SCHEMAS[key] = compiled_schema
        compiled_schema(object_)
        return ParseResult(object=object_, resource=None, exception=None, path=None, resource_id=object_.get('id', None))
    except (ValidationError, AssertionError) as e:
        if object_:
            return ParseResult(object=object_, resource=None, exception=e, path=None, resource_id=object_.get('id', None))
        else:
            return ParseResult(object=object_, resource=None, exception=e, path=None)


def directory_reader(
        directory_path: pathlib.Path,
        schema_path: str,
        pattern: str = '*.ndjson',
        validate_=True) -> Iterator[ParseResult]:
    """Extract FHIR resources from directory"""

    assert directory_path.is_dir(), f"{directory_path.name} is not a directory"

    schemas = ensure_schema(schema_path)

    input_files = [_ for _ in directory_path.glob(pattern)]
    for input_file in input_files:
        logger.info(input_file)
        if not input_file.is_file():
            continue
        fp = _to_file(input_file)
        with fp:
            offset = 0
            for line in fp.readlines():
                gen3_resource = orjson.loads(line)
                parse_result = validate(gen3_resource, schemas=schemas)
                parse_result.path = input_file
                parse_result.offset = offset
                offset += 1
                yield parse_result


def ensure_schema(schema_path):
    """Ensure schema is loaded, from either local file or URL."""
    if 'http' in schema_path:
        schemas = DataDictionary(url=schema_path).schema
    elif os.path.exists(schema_path):
        schemas = DataDictionary(local_file=schema_path).schema
    return schemas



import fastjsonschema
from iceberg_tools.schema.simplified_validator import ensure_schema


def test_schema(distribution_schema):
    """Ensure the schema compiles as a legal json schema."""
    compiled_schema = fastjsonschema.compile(distribution_schema)
    assert compiled_schema


def test_schema_load(distribution_schema_path):
    """Ensure we can load using dictionaryutils."""
    assert ensure_schema(schema_path=distribution_schema_path), f"should have loaded {distribution_schema_path}"


def test_schema_denormalized_fields(distribution_schema_path):
    """Ensure specimen and document_reference have _identifier denormalized fields."""
    schema = ensure_schema(schema_path=distribution_schema_path)
    assert schema, f"should have loaded {distribution_schema_path}"

    properties = schema['specimen']['properties']
    for _ in ['patient_identifier']:
        assert _ in properties, f"should have denormalized field {_}"

    properties = schema['document_reference']['properties']
    for _ in ['patient_identifier', 'specimen_identifier', 'task_identifier']:
        assert _ in properties, f"should have denormalized field {_}"

    properties = schema['observation']['properties']
    for _ in ['patient_identifier', 'specimen_identifier', 'condition_identifier']:
        assert _ in properties, f"should have denormalized field {_}"



import pathlib
import yaml
from jsonschema.validators import Draft202012Validator


def test_yaml_anonymous_schemas(bmeg_dir):
    """Check individual anonymous schemas"""
    for file_name in pathlib.Path('').glob("*.yaml"):
        with open(file_name) as fp:
            schema = yaml.load(fp, yaml.SafeLoader)
            Draft202012Validator.check_schema(schema)
            description = schema.get('description', None)
            assert description is not None, f"Missing description in {file_name}"
            assert '[See' in description, f"Missing [See ...] in {file_name}"
            # test to ensure See only appears once
            assert description.count('[See') == 1, f"Multiple [See ...] in {file_name}"



from .errors import DictionaryError
from .version_data import DICTVERSION, DICTCOMMIT

from copy import deepcopy
from collections import namedtuple
from contextlib import contextmanager
from jsonschema import RefResolver

import glob
import logging
import requests
import os
import yaml
import json

MOD_DIR = os.path.abspath(os.path.dirname(__file__))

ResolverPair = namedtuple("ResolverPair", ["resolver", "source"])


@contextmanager
def visit_directory(path):
    """Perform contained actions with current working directory at
    :param:``path``.  Always return to previous directory when done.

    """

    cdir = os.getcwd()
    try:
        os.chdir(path)
        yield os.getcwd()
    finally:
        os.chdir(cdir)


def load_yaml(name):
    """Return contents of yaml file as dict"""
    with open(name, "r", encoding="utf8") as f:
        return yaml.safe_load(f)


def add_default_schema(dictionary):
    """
    add default schema nodes to a dictionary
    """
    yamls, resolvers = load_schemas_from_dir(os.path.join(MOD_DIR, "schemas"))
    dictionary.resolvers.update(resolvers)

    schemas = {
        schema["id"]: dictionary.resolve_schema(schema, deepcopy(schema))
        for path, schema in yamls.items()
        if path not in dictionary.exclude
    }
    dictionary.schema.update(schemas)


def load_schemas_from_url(url, logger, schemas=None, resolvers=None):
    error_msg = "Fail to get schema from {}".format(url)
    try:
        r = requests.get(url)
    except Exception as e:
        error_msg = "{}: {}".format(
            error_msg, e.message if hasattr(e, "message") else e
        )
        logger.exception(error_msg)
        raise DictionaryError(error_msg)
    if r.status_code != 200:
        error_msg = "{}: {}".format(error_msg, r.text)
        logger.error(error_msg)
        raise DictionaryError(error_msg)
    if schemas is None:
        schemas = {}
    if resolvers is None:
        resolvers = {}
    response = json.loads(r.text)
    for key, schema in response.items():
        schemas[key] = schema
        resolver = RefResolver("{}#".format(key), schema)
        resolvers[key] = ResolverPair(resolver, schema)

    return schemas, resolvers


def load_schemas_from_file(local_file, schemas=None, resolvers=None):
    if schemas is None:
        schemas = {}
    if resolvers is None:
        resolvers = {}
    with open(local_file, "r") as r:
        response = json.loads(r.read())
        for key, schema in response.items():
            schemas[key] = schema
            resolver = RefResolver("{}#".format(key), schema)
            resolvers[key] = ResolverPair(resolver, schema)
    return schemas, resolvers


def dump_schemas_from_dir(directory):
    """Dump schema as a json"""

    with visit_directory(directory):
        result = {path: load_yaml(path) for path in glob.glob("*.yaml")}
        if not ("_settings.yaml" in result):
            result["_settings.yaml"] = {}
        result["_settings.yaml"]["_dict_commit"] = DICTCOMMIT
        result["_settings.yaml"]["_dict_version"] = DICTVERSION
        return result


def load_schemas_from_dir(directory, schemas=None, resolvers=None):
    """Returns all yamls and resolvers of those yamls from dir"""

    if schemas is None:
        schemas = {}
    if resolvers is None:
        resolvers = {}

    with visit_directory(directory):
        for path in glob.glob("*.yaml"):
            schema = load_yaml(path)
            schemas[path] = schema
            resolver = RefResolver("{}#".format(path), schema)
            resolvers[path] = ResolverPair(resolver, schema)

    return schemas, resolvers


class DataDictionary(object):

    _metaschema_path = "metaschema.yaml"
    _definitions_paths = ["_definitions.yaml", "_terms.yaml"]
    settings_path = "_settings.yaml"

    logger = logging.getLogger("DataDictionary")

    def __init__(
        self,
        lazy=False,
        root_dir=None,
        definitions_paths=None,
        metaschema_path=None,
        url=None,
        local_file=None,
    ):
        """Creates a new dictionary instance.

        :param root_dir: The directory to find schemas
        :param metaschema_path: The metaschema to validate schemas with
        :param definitions_paths: Paths to resolve $ref to
        :param lazy: If true, wait to load dictionary
        :param url:Load data from an url

        """

        self.root_dir = root_dir or os.path.join(MOD_DIR, "schemas")
        self.metaschema_path = metaschema_path or self._metaschema_path
        self.definitions_paths = definitions_paths or self._definitions_paths
        self.exclude = (
            [self.metaschema_path] + self.definitions_paths + [self.settings_path]
        )
        self.schema = dict()
        self.resolvers = dict()

        self.metaschema = load_yaml(
            os.path.join(MOD_DIR, "schemas", self.metaschema_path)
        )

        if not lazy:
            self.load_data(directory=self.root_dir, url=url, local_file=local_file)
        self.allow_nulls()

    def load_data(self, directory=None, url=None, local_file=None):
        """Load and reslove all schemas from directory or url"""
        yamls, resolvers = load_schemas_from_dir(os.path.join(MOD_DIR, "schemas"))
        if url:
            yamls, resolvers = load_schemas_from_url(
                url, self.logger, schemas=yamls, resolvers=resolvers
            )
        elif local_file:
            yamls, resolvers = load_schemas_from_file(
                local_file, schemas=yamls, resolvers=resolvers
            )
        else:
            yamls, resolvers = load_schemas_from_dir(
                directory, schemas=yamls, resolvers=resolvers
            )

        self.settings = yamls.get(self.settings_path) or {}
        self.resolvers.update(resolvers)

        schemas = {
            schema["id"]: self.resolve_schema(schema, deepcopy(schema))
            for path, schema in yamls.items()
            if path not in self.exclude
        }
        self.schema.update(schemas)

    def resolve_reference(self, value, root):
        """Resolves a reference.

        :param value: The actual reference, e.g. ``_yaml.yaml#/def``
        :param root:
            The containing root of :param:`value`. This needs to be
            passed in order to resolve self-referential $refs,
            e.g. ``#/def``.
        :returns: JSON Schema pointed to by :param:`value`

        """
        base, ref = value.split("#", 1)

        if base:
            resolver, new_root = self.resolvers[base]
            referrer, resolution = resolver.resolve(value)
            self.resolve_schema(resolution, new_root)
        else:
            resolver = RefResolver("#", root)
            referrer, resolution = resolver.resolve(value)

        return resolution

    def resolve_schema(self, obj, root):
        """Recursively resolves all references in a schema against
        ``self.resolvers``.

        :param obj: The object to recursively resolve.
        :param root:
            The containing root of :param:`value`. This needs to be
            passed in order to resolve self-referential $refs,
            e.g. ``#/def``.
        :returns: A denormalized/resolved version of :param:`obj`.

        """
        refkey = "$ref"
        if isinstance(obj, dict):
            all_refkeys = [k for k in obj.keys() if k.startswith(refkey)]
            for key in all_refkeys:
                val = obj.pop(key)
                obj.update(self.resolve_reference(val, root))
            return {k: self.resolve_schema(v, root) for k, v in obj.items()}
        elif isinstance(obj, list):
            return [self.resolve_schema(item, root) for item in obj]
        else:
            return obj

    def get_required_link_names(self, links):
        required_links = []
        for link in links:
            if "subgroup" in link:
                required_links.extend(self.get_required_link_names(link["subgroup"]))
            else:
                if link["required"]:
                    required_links.append(link["name"])
        return required_links

    def allow_nulls(self):
        """
        Adds "none" to the types of non-required fields in the dictionary.
        This is done so we can remove properties/links from entities by
        updating the property to null.
        """
        for node_properties in self.schema.values():
            required = node_properties.get("required", []).copy()
            required.extend(
                self.get_required_link_names(node_properties.get("links", []))
            )

            for prop_id, prop in node_properties.get("properties", {}).items():
                if prop_id in required:
                    continue

                if "type" in prop and "null" not in prop["type"]:
                    # add null type to optional props and links
                    if not isinstance(prop["type"], list):
                        prop["type"] = [prop["type"]]
                    prop["type"].append("null")
                elif "enum" in prop and None not in prop["enum"]:
                    # add null type to optional enums
                    if not isinstance(prop["enum"], list):
                        prop["enum"] = [prop["enum"]]
                    prop["enum"].append(None)
                elif "anyOf" in prop and {"type": "null"} not in prop["anyOf"]:
                    prop["anyOf"].append({"type": "null"})
                elif "oneOf" in prop and {"type": "null"} not in prop["oneOf"]:
                    prop["oneOf"].append({"type": "null"})
